"""
Parsers for inputing experimental files.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import pandas as pd


def i07_dat_parser(file_path):
    """
    Parsing the .dat file from I07.

    Args:
        (:py:attr:`str`): The ``.dat`` file to be read.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`dict`: The metadata from the ``.dat`` file.
            - :py:class:`pandas.DataFrame`: The data from the ``.dat`` file.
    """
    f_open = open(file_path, "r")
    # Neither the data nor the metadata are being read yet.
    data_reading = False
    metadata_reading = False

    # Create the dictionaries to be populated.
    data_dict = {}
    metadata_dict = {}
    # Create the list to be filled with lists for each line
    data_lines = []

    for line in f_open:
        # This string incidates the start of the metadata.
        if "<MetaDataAtStart>" in line:
            metadata_reading = True
        # This string indicates the end of the metadata.
        if "</MetaDataAtStart>" in line:
            metadata_reading = False
        # This string indicates the start of the data.
        if " &END" in line:
            data_reading = True
            # Set counter to minus two, such that when is
            # reaches the data it is 0.
            count = -2
        # When the metadata section is being read populate the metadata_dict
        if metadata_reading:
            if "=" in line:
                metadata_in_line = []
                for i in line.split("=")[1:]:
                    try:
                        j = float(i)
                    except ValueError:
                        j = i
                    metadata_in_line.append(j)
                metadata_dict[line.split("=")[0]] = metadata_in_line
        # When the data section is being read, make the list of the zeroth line
        # the titles and everything after is the data_lines list of lists.
        if data_reading:
            count += 1
            if count == 0:
                titles = line.split()
            if count > 0:
                data_lines.append(line.split())
    f_open.close()
    # Sort the data_lines list of lists to transpore and make into a dict where
    # the keys are the titles.
    for j in range(len(data_lines[0])):
        list_to_add = []
        for i in range(len(data_lines)):
            try:
                list_to_add.append(float(data_lines[i][j]))
            except ValueError:
                list_to_add.append(data_lines[i][j])
        count = 0
        if j >= len(titles):
            data_dict[str(count)] = list_to_add
            count += 1
        else:
            data_dict[titles[j]] = list_to_add
    return metadata_dict, pd.DataFrame(data_dict)



"""
Tests for corrections module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)
# pylint: disable=R0201

import unittest
import numpy as np
from numpy.testing import assert_almost_equal
from islatu import cropping


class TestCropping(unittest.TestCase):
    """
    Unit tests for cropping module
    """

    def test_crop_2d_a(self):
        """
        Test crop_2d.
        """
        initial_array = np.ones((50, 50))
        expected_array = np.ones((20, 20))
        result = cropping.crop_2d(initial_array, None, 20, 40, 10, 30)
        assert_almost_equal(result, expected_array)

    def test_crop_2d_b(self):
        """
        Test crop_2d with rectanglar crop.
        """
        initial_array = np.ones((50, 50))
        expected_array = np.ones((10, 20))
        result = cropping.crop_2d(initial_array, None, 20, 40, 10, 20)
        assert_almost_equal(result, expected_array)

    def test_crop_2d_b(self):
        """
        Test crop_2d with rectanglar crop.
        """
        initial_array = np.ones((50, 50)) * 10
        initial_array_e = np.ones((50, 50))
        expected_array = np.ones((10, 20)) * 10
        expected_array_e = np.ones((10, 20))
        result, result_e = cropping.crop_2d(initial_array, initial_array_e, 20, 40, 10, 20)
        assert_almost_equal(result, expected_array)
        assert_almost_equal(result_e, expected_array_e)

    def test_crop_around_peak_2d_a(self):
        """
        Test crop_around_peak_2d with custom.
        """
        initial_array = np.ones((50, 50))
        initial_array[25, 25] = 100
        expected_array = np.ones((10, 10))
        expected_array[5, 5] = 100
        result = cropping.crop_around_peak_2d(initial_array, None, x_size=10, y_size=10)
        assert_almost_equal(result, expected_array)

    def test_crop_around_peak_2d_b(self):
        """
        Test crop_around_peak_2d with defaults.
        """
        initial_array = np.ones((50, 50))
        initial_array[25, 25] = 100
        expected_array = np.ones((20, 20))
        expected_array[10, 10] = 100
        result = cropping.crop_around_peak_2d(initial_array, None)
        assert_almost_equal(result, expected_array)

    def test_crop_around_peak_2d_c(self):
        """
        Test crop_around_peak_2d with asymmetry.
        """
        initial_array = np.ones((50, 50))
        initial_array[25, 25] = 100
        expected_array = np.ones((20, 10))
        expected_array[10, 5] = 100
        result = cropping.crop_around_peak_2d(initial_array, None, x_size=10, y_size=20)
        assert_almost_equal(result, expected_array)

    def test_crop_around_peak_2d_error(self):
        """
        Test crop_around_peak_2d with uncertainty.
        """
        initial_array = np.ones((50, 50))
        initial_array[25, 25] = 100
        initial_array_e = initial_array * 0.1
        expected_array = np.ones((10, 10))
        expected_array[5, 5] = 100
        expected_array_e = expected_array * 0.1
        result = cropping.crop_around_peak_2d(
            initial_array, initial_array_e, x_size=10, y_size=10
        )
        assert_almost_equal(result[0], expected_array)
        assert_almost_equal(result[1], expected_array_e)



"""
The refl_data module includes two classes,
:py:class:`~islatu.refl_data.Profile` and :py:class:`~islatu.refl_data.Scan`
that are integral to the use of :py:mod`islatu` for the reduction of x-ray
reflectometry data.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from os import path
import numpy as np
from scipy.constants import physical_constants
from scipy.interpolate import splev
from uncertainties import ufloat
from uncertainties import unumpy as unp
from islatu import corrections, image, stitching


class Profile:
    """
    This class stores information about the reflectometry profile.

    Attributes:
        scans (:py:attr:`list` of :py:class:`islatu.refl_data.Scan`):
            Reflectometry scans that make up the profile.
        q_vectors (:py:attr:`array_like`): q-vectors from the reflectometry
            scan, populated after concatentation.
        reflected_intensity (:py:attr:`array_like`): Reflected intensities
            from the reflectometry scan, populated after concatentation.

    Args:
        file_paths (:py:attr:`list`): List of files, one for each
            reflectometry scan.
        parser (:py:attr:`callable`): Parser function for the reflectometry
            scan files.
        q_axis_name (:py:attr:`str`, optional): Label for the q-axis in the
            scan. Defaults to :py:attr:`'q_axis_name'`.
        theta_axis_name (:py:attr:`str`, optional): Label for the theta axis
            in the scan. Defaults to :py:attr:`'dcdtheta'`.
        energy (:py:attr:`float`): The energy of the probing X-ray, in keV.
            Defaults to finding from metadata if no value given.
        pixel_min (:py:attr:`int`): The minimum pixel value possible, all
            pixels found with less than this value will have this value
            assigned. Defaults to :py:attr:`0`.
        pixel_max (:py:attr:`int`): The maximum pixel value allowed, if any
            detector image has a pixel over this value, the whole image is
            removed from the dataset. This is due to the risk of non-linear
            counting at very high counts. Defaults to :py:attr:`1e6`.
        hot_pixel_max (:py:attr:`int`): The value of a hot pixel that will be
            set to the average of the surrounding pixels. Logically, this
            should be less than the :py:attr:`pixel_max` value. Defaults to
            :py:attr:`2e5`.
        progress (:py:attr:`bool`, optional): Show a progress bar.
            Requires the :py:mod:`tqdm` package. Defaults to :py:attr:`True`.
    """
    def __init__(self, file_paths, parser, q_axis_name="qdcd",
                 theta_axis_name="dcdtheta", energy=None, pixel_min=0,
                 pixel_max=1000000, hot_pixel_max=1e5, transpose=False):
        self.scans = []
        try:
            _ = len(pixel_min)
        except TypeError:
            pixel_min = [pixel_min] * len(file_paths)
        try:
            _ = len(pixel_max)
        except TypeError:
            pixel_max = [pixel_max] * len(file_paths)
        try:
            _ = len(hot_pixel_max)
        except TypeError:
            hot_pixel_max = [hot_pixel_max] * len(file_paths)
        for i, fyle in enumerate(file_paths):
            self.scans.append(
                Scan(fyle, parser, q_axis_name, theta_axis_name,
                     pixel_min=pixel_min[i], pixel_max=pixel_max[i], hot_pixel_max=hot_pixel_max[i], transpose=transpose))
        self.q_vectors = None
        self.reflected_intensity = None

    @property
    def R(self):
        """
        Reflected intensity values.

        Returns:
            :py:attr:`array_like`: Intensity values.
        """
        return unp.nominal_values(self.reflected_intensity)

    @property
    def dR(self):
        """
        Reflected intensity uncertainties.

        Returns:
            :py:attr:`array_like`: Intensity uncertainties.
        """
        return unp.std_devs(self.reflected_intensity)

    @property
    def q(self):
        """
        q-value values.

        Returns:
            :py:attr:`array_like`: q-value values.
        """
        return unp.nominal_values(self.q_vectors)

    @property
    def dq(self):
        """
        q-value uncertainties.

        Returns:
            :py:attr:`array_like`: q-value uncertainties.
        """
        return unp.std_devs(self.q_vectors)

    def crop(self, crop_function, kwargs=None, progress=True):
        """
        Class method for the :func:`~islatu.refl_data.Scan.crop` method for
        each :py:class:`~Scan` in the list.

        Args:
            crop_function (:py:attr:`callable`): Cropping function to be used.
            kwargs (:py:attr:`dict`, optional): Keyword arguments for the
                cropping function. Defaults to :py:attr:`None`.
            progress (:py:attr:`bool`, optional): Show a progress bar.
                Requires the :py:mod:`tqdm` package. Defaults to
                :py:attr:`True`.
        """
        for scan in self.scans:
            scan.crop(crop_function, kwargs, progress)

    def bkg_sub(self, bkg_sub_function, kwargs=None, progress=True):
        """
        Class method for the :func:`~islatu.refl_data.Scan.bkg_sub` method for
        each :py:class:`~Scan` in the list.

        Args:
            bkg_sub_function (:py:attr:`callable`): Background subtraction
                function to be used.
            kwargs (:py:attr:`dict`, optional): Keyword arguments for
                the background subtraction function. Defaults to
                :py:attr:`None`.
            progress (:py:attr:`bool`, optional): Show a progress bar.
                Requires the :py:mod:`tqdm` package. Defaults to
                :py:attr:`True`.
        """
        for scan in self.scans:
            scan.bkg_sub(bkg_sub_function, kwargs, progress)

    def footprint_correction(self, beam_width, sample_size):
        """
        Class method for :func:`~islatu.refl_data.Scan.footprint_correction`
        for each :py:class:`~Scan` in the list.

        Args:
            beam_width (:py:attr:`float`): Width of incident beam, in metres.
            sample_size (:py:class:`uncertainties.core.Variable`): Width of
                sample in the dimension of the beam, in metres.
            theta (:py:attr:`float`): Incident angle, in degrees.
        """
        for scan in self.scans:
            scan.footprint_correction(beam_width, sample_size)

    def transmission_normalisation(self):
        """
        Perform the transmission correction
        (:func:`~islatu.refl_data.Scan.transmission_normalisation`) for each
        :py:class:`~Scan` in the list and fine attenutation correction
        (:func:`~islatu.stitching.correction_attenutation`).
        """
        for scan in self.scans:
            scan.transmission_normalisation()
        self.scans = stitching.correct_attentuation(self.scans)

    def resolution_function(self, qz_dimension=1, progress=True,
                                 detector_distance=None, energy=None,
                                 pixel_size=172e-6):
        """
        Class method for
        :func:`~islatu.refl_data.Scan.q_uncertainty_from_pixel` for each
        :py:class:`~Scan` in the list.

       Args:
           qz_dimension (:py:attr:`int`, optional): The dimension of q_z in
               the detector image (this should be the opposite index to that
               the summation is performed if the
               :py:func:`islatu.background.fit_gaussian_1d` background
               subtraction has been performed). Defaults to :py:attr:`1`.
            number_of_pixels (:py:attr:`float`): Number of pixels of q
                uncertainty.
            detector_distance (:py:attr:`float`): Sample detector distance in
                metres
            energy (:py:attr:`float`): X-ray energy in keV
            pixel_size (:py:attr:`float`, optional): Pixel size in metres
        """
        for scan in self.scans:
            scan.resolution_function(
                qz_dimension, progress, detector_distance,
                energy, pixel_size
            )

    def qdcd_normalisation(self, itp):
        """
        Class method for :func:`~islatu.refl_data.Scan.qdcd_normalisation` for
        each :py:class:`~Scan` in the list.

        Args:
            normalisation_file (:py:attr:`str`): The ``.dat`` file that
                contains the normalisation data.
        """
        for scan in self.scans:
            scan.qdcd_normalisation(itp)

    def concatenate(self):
        """
        Class method for :func:`~islatu.stitching.concatenate`.
        """
        self.q_vectors, self.reflected_intensity = stitching.concatenate(
            self.scans)

    def normalise_ter(self, max_q=0.1):
        """
        Class method for :func:`~islatu.stitching.normalise_ter`.

        Args:
            max_q (:py:attr:`float`): The maximum q to be included in finding
                the critical angle.
        """
        self.reflected_intensity = stitching.normalise_ter(
            self.q_vectors, self.reflected_intensity, max_q
        )

    def rebin(self, new_q=None, number_of_q_vectors=400, interpolate=False):
        """
        Class method for :func:`islatu.stitching.rebin`.

        Args:
            new_q (:py:attr:`array_like`): Array of potential q-values.
                Defaults to :py:attr:`None`.
            number_of_q_vectors (:py:attr:`int`, optional): The max number of
                q-vectors to be using initially in the rebinning of the data.
                Defaults to :py:attr:`400`.
        """
        self.q_vectors, self.reflected_intensity = stitching.rebin(
            self.q_vectors, self.reflected_intensity,
            new_q, number_of_q_vectors, interpolate)


class Scan:
    """
    This class stores scan specific information.

    Attributes:
        file_paths (:py:attr:`str`): File path for scan file.
        metadata (:py:attr:`dict`): The metadata from the ``.dat`` file.
        data (:py:class:`pandas.DataFrame`): The data from the ``.dat`` file.
        images (:py:attr:`list` of :py:class:`islatu.image.Image`): The
            detector images in the given scan.
        q (:py:attr:`array_like`): The measured q-values in the scan.
        theta (:py:attr:`array_like`): The measured theta values in the scan.
        R (:py:attr:`array_like`): The reflected intensity values in the scan.
        n_pixels (:py:attr:`array_like`): The width in pixels of the beam in
            each image.

    Args:
        file_paths (:py:attr:`list`): List of files, one for each
            reflectometry scan.
        parser (:py:attr:`callable`): Parser function for the reflectometry
            scan files.
        q_axis_name (:py:attr:`str`, optional): Label for the q-axis in the
            scan. Defaults to :py:attr:`'q_axis_name'`.
        theta_axis_name (:py:attr:`str`, optional): Label for the theta axis
            in the scan. Defaults to :py:attr:`'dcdtheta'`.
        energy (:py:attr:`float`): The energy of the probing X-ray, in keV.
            Defaults to finding from metadata if no value given.
        pixel_min (:py:attr:`int`): The minimum pixel value possible, all
            pixels found with less than this value will have this value
            assigned. Defaults to :py:attr:`0`.
        pixel_max (:py:attr:`int`): The maximum pixel value allowed, if any
            detector image has a pixel over this value, the whole image is
            removed from the dataset. This is due to the risk of non-linear
            counting at very high counts. Defaults to :py:attr:`1e6`.
        hot_pixel_max (:py:attr:`int`): The value of a hot pixel that will be
            set to the average of the surrounding pixels. Logically, this
            should be less than the :py:attr:`pixel_max` value. Defaults to
            :py:attr:`2e5`.
        progress (:py:attr:`bool`, optional): Show a progress bar.
            Requires the :py:mod:`tqdm` package. Defaults to :py:attr:`True`.
    """

    def __init__(self, file_path, parser, q_axis_name="qdcd",
                 theta_axis_name="dcdtheta", energy=None, pixel_min=0,
                 pixel_max=1000000, hot_pixel_max=2e5,
                 transpose=False, progress=True):
        self.file_path = file_path
        self.metadata, self.data = parser(self.file_path)
        if q_axis_name is None:
            planck = physical_constants["Planck constant in eV s"][0] * 1e-3
            speed_of_light = physical_constants[
                "speed of light in vacuum"][0] * 1e10
            if energy is None:
                energy = self.metadata["dcm1energy"][0]
            q_values = energy * 4 * np.pi * unp.sin(
                unp.radians(
                    self.data[theta_axis_name])) / (planck * speed_of_light)
            self.q = unp.uarray(
                q_values, np.zeros(self.data[theta_axis_name].size))
        else:
            self.q = unp.uarray(
                self.data[q_axis_name], np.zeros(self.data[q_axis_name].size)
            )
        self.data = self._check_files_exist()
        self.theta = unp.uarray(
            self.data[theta_axis_name],
            np.zeros(self.data[theta_axis_name].size))
        self.R = unp.uarray(np.zeros(self.q.size), np.zeros(self.q.size),)
        self.n_pixels = np.zeros(self.q.size)
        self.transpose = transpose
        self.images = []
        iterator = _get_iterator(unp.nominal_values(self.q), progress)
        to_remove = []
        for i in iterator:
            if self.data['roi1_maxval'][i] <= pixel_max:
                img = image.Image(self.data["file"][i], self.data,
                                  self.metadata, pixel_min=pixel_min,
                                  hot_pixel_max=hot_pixel_max,
                                  transpose=self.transpose)
                self.images.append(img)
            else:
                to_remove.append(i)
        self.R = np.delete(self.R, to_remove)
        self.theta = np.delete(self.theta, to_remove)
        self.n_pixels = np.delete(self.n_pixels, to_remove)
        self.q = np.delete(self.q, to_remove)


    def __str__(self):
        """
        Custom string output.

        Returns:
            :py:attr:`str`: Description of scan.
        """
        return ("The file: {} contains {} images from q = {:.4f} to "
                "q = {:.4f}.".format(self.file_path, self.q.size,
                                     self.q[0].n, self.q[-1].n))

    def __repr__(self):
        """
        Custom repr output.

        Returns:
            :py:attr:`str`: Description of scan.
        """
        return self.__str__()

    def _check_files_exist(self):
        """
        Check that image files exist.

        Returns:
            :py:class:`pandas.DataFrame`: Modified data with corrected file paths.
        """
        iterator = _get_iterator(unp.nominal_values(self.q), False)
        for i in iterator:
            im_file = self.data["file"][i]
            if path.isfile(im_file):
                continue
            im_file = self.data["file"][i].split('/')[-2:]
            im_file = path.join(im_file[0], im_file[1])
            im_file = path.join(path.dirname(self.file_path), im_file)
            if path.isfile(im_file):
                self.data.iloc[
                    i, self.data.keys().get_loc("file")] = im_file
                continue
            im_file = self.data["file"][i].split('/')[-1]
            im_file = path.join(path.dirname(self.file_path), im_file)
            if path.isfile(im_file):
                self.data.iloc[
                    i, self.data.keys().get_loc("file")] = im_file
                continue
            raise FileNotFoundError(
                "The following image file could not be "
                "found: {}.".format(self.data["file"][i]))
        return self.data

    def crop(self, crop_function, kwargs=None, progress=True):
        """
        Perform detector image cropping.

        args:
            crop_function (:py:attr:`callable`): Cropping function to be used.
            kwargs (:py:attr:`dict`, optional): Keyword arguments for the
                cropping function. Defaults to :py:attr:`None`.
            progress (:py:attr:`bool`, optional): Show a progress bar.
                Requires the :py:mod:`tqdm` package. Defaults
                to :py:attr:`True`.
        """
        iterator = _get_iterator(self.images, progress)
        for i in iterator:
            if kwargs is None:
                self.images[i].crop(crop_function)
            else:
                self.images[i].crop(crop_function, **kwargs)
            self.R[i] = ufloat(self.images[i].sum().n, self.images[i].sum().s)

    def bkg_sub(self, bkg_sub_function, kwargs=None, progress=True):
        """
        Perform background substraction for each image in a Scan.

        Args:
            bkg_sub_function (:py:attr:`callable`): Background subtraction
                function to be used.
            kwargs (:py:attr:`dict`, optional): Keyword arguments for
                the background subtraction function. Defaults
                to :py:attr:`None`.
            progress (:py:attr:`bool`, optional): Show a progress bar.
                Requires the :py:mod:`tqdm` package. Defaults
                to :py:attr:`True`.
        """
        iterator = _get_iterator(self.images, progress)
        for i in iterator:
            if kwargs is None:
                self.images[i].background_subtraction(bkg_sub_function)
            else:
                self.images[i].background_subtraction(
                    bkg_sub_function, **kwargs)
            self.R[i] = ufloat(self.images[i].sum().n, self.images[i].sum().s)

    def resolution_function(self, qz_dimension=1, progress=True,
                            detector_distance=None, energy=None,
                            pixel_size=172e-6):
        """
        Estimate the q-resolution function based on the reflected intensity
        on the detector and add this to the q uncertainty.

        Args:
            qz_dimension (:py:attr:`int`, optional): The dimension of q_z in
                the detector image (this should be the opposite index to that
                the summation is performed if the
                :py:func:`islatu.background.fit_gaussian_1d` background
                subtraction has been performed). Defaults to :py:attr:`1`.
            progress (:py:attr:`bool`, optional): Show a progress bar.
                Requires the :py:mod:`tqdm` package. Defaults
                to :py:attr:`True`.
            detector_distance (:py:attr:`float`): Sample detector distance in
                metres
            energy (:py:attr:`float`): X-ray energy in keV
            pixel_size (:py:attr:`float`, optional): Pixel size in metres
        """
        iterator = _get_iterator(self.images, progress)
        for i in iterator:
            self.images[i].q_resolution(qz_dimension)
        if detector_distance is None:
            detector_distance = self.metadata["diff1detdist"][0] * 1e-3
        if energy is None:
            energy = self.metadata["dcm1energy"][0]
        offset = np.arctan(
            pixel_size * 1.96 * self.n_pixels * 0.5 / (detector_distance)
        )
        planck = physical_constants["Planck constant in eV s"][0] * 1e-3
        speed_of_light = physical_constants[
            "speed of light in vacuum"][0] * 1e10
        q_uncertainty = energy * 4 * np.pi * unp.sin(
            offset) / (planck * speed_of_light)
        self.q = unp.uarray(
            unp.nominal_values(self.q), unp.std_devs(self.q) + q_uncertainty)

    def footprint_correction(self, beam_width, sample_size):
        """
        Class method for :func:`islatu.corrections.footprint_correction`.

        Args:
            beam_width (:py:attr:`float`): Width of incident beam, in metres.
            sample_size (:py:class:`uncertainties.core.Variable`): Width of
                sample in the dimension of the beam, in metres.
            theta (:py:attr:`float`): Incident angle, in degrees.
        """
        self.R /= corrections.footprint_correction(
            beam_width, sample_size, self.theta)

    def transmission_normalisation(self):
        """
        Perform the transmission correction.
        """
        self.R /= float(self.metadata["transmission"][0])

    def qdcd_normalisation(self, itp):
        """
        Perform normalisation by DCD variance.

        Args:
            itp (:py:attr:`tuple`): Containing interpolation knots
                (:py:attr:`array_like`), B-spline coefficients
                (:py:attr:`array_like`), and degree of spline (:py:attr:`int`).
        """
        self.R /= splev(unp.nominal_values(self.q), itp)


def _get_iterator(to_iter, progress):
    """
    Create an iterator.

    Args:
        to_iter (:py:attr:`array_like`): The list or array to iterate.
        progress (:py:attr:`bool`): Show progress bar.

    Returns:
        :py:attr:`range` or :py:class:`tqdm.std.tqdm`: Iterator object.
    """
    iterator = range(len(to_iter))
    if progress:
        try:
            from tqdm import tqdm
            iterator = tqdm(range(len(to_iter)))
        except ModuleNotFoundError:
            print(
                "For the progress bar, you need to have the tqdm package "
                "installed. No progress bar will be shown"
            )
    return iterator



"""
Tests for refl_data module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)

import unittest
from os import path
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
import islatu
from islatu import stitching, refl_data, io
from uncertainties import unumpy


class TestStitcher(unittest.TestCase):
    """
    Unit tests for stitching module
    """

    def test_correct_attentuation(self):
        """
        Test correction of attenutation
        """
        y1 = np.ones(10) * 10
        dy1 = np.ones(10)
        y2 = np.ones(10) * 5
        dy2 = np.ones(10) * 0.5

        x1 = np.arange(1, 11, 1)
        dx1 = np.copy(x1) * 0.05
        x2 = np.arange(8, 18, 1)
        dx2 = np.copy(x2) * 0.05

        scan1 = refl_data.Scan(
            path.join(path.dirname(islatu.__file__), "tests/test_files/test_a.dat"),
            io.i07_dat_parser,
        )
        scan2 = refl_data.Scan(
            path.join(path.dirname(islatu.__file__), "tests/test_files/test_a.dat"),
            io.i07_dat_parser,
        )

        scan1.q = unumpy.uarray(x1, dx1)
        scan1.R = unumpy.uarray(y1, dy1)
        scan2.q = unumpy.uarray(x2, dx2)
        scan2.R = unumpy.uarray(y2, dy2)
        scan_list = [scan1, scan2]

        exp_y1 = np.ones(10) * 10
        exp_dy1 = np.ones(10)
        exp_y2 = np.ones(10) * 10
        exp_dy2 = np.append(np.ones(3) * 0.8164966, np.ones(7) * 1.1547005)

        scan_list = stitching.correct_attentuation(scan_list)

        assert_almost_equal(unumpy.nominal_values(scan_list[0].R), exp_y1)
        assert_almost_equal(unumpy.std_devs(scan_list[0].R), exp_dy1)
        assert_almost_equal(unumpy.nominal_values(scan_list[1].R), exp_y2)
        assert_almost_equal(unumpy.std_devs(scan_list[1].R), exp_dy2)

    def test_concatenate(self):
        """
        Test concatenation
        """
        y1 = np.ones(10) * 10
        dy1 = np.ones(10)
        y2 = np.ones(10) * 5
        dy2 = np.ones(10) * 0.5

        x1 = np.arange(1, 11, 1)
        dx1 = np.copy(x1) * 0.05
        x2 = np.arange(8, 18, 1)
        dx2 = np.copy(x2) * 0.05

        scan1 = refl_data.Scan(
            path.join(path.dirname(islatu.__file__), "tests/test_files/test_a.dat"),
            io.i07_dat_parser,
        )
        scan2 = refl_data.Scan(
            path.join(path.dirname(islatu.__file__), "tests/test_files/test_a.dat"),
            io.i07_dat_parser,
        )

        scan1.q = unumpy.uarray(x1, dx1)
        scan1.R = unumpy.uarray(y1, dy1)
        scan2.q = unumpy.uarray(x2, dx2)
        scan2.R = unumpy.uarray(y2, dy2)
        scan_list = [scan1, scan2]

        exp_y1 = unumpy.uarray(np.append(y1, y2), np.append(dy1, dy2))
        exp_x1 = unumpy.uarray(np.append(x1, x2), np.append(dx1, dx2))

        q, r = stitching.concatenate(scan_list)

        assert_almost_equal(unumpy.nominal_values(r), unumpy.nominal_values(exp_y1))
        assert_almost_equal(unumpy.std_devs(r), unumpy.std_devs(exp_y1))
        assert_almost_equal(unumpy.nominal_values(q), unumpy.nominal_values(exp_x1))
        assert_almost_equal(unumpy.std_devs(q), unumpy.std_devs(exp_x1))

    def test_normalise_ter(self):
        """
        Test normalisation to 1
        """
        y = [1e9, 1e9, 1e9, 3.16e8, 1.29e8, 6.25e7, 3.37e7, 2.00e7]
        dy = [1e7, 1e7, 1e7, 3.16e6, 1.29e6, 6.25e5, 3.37e5, 2.00e5]

        x = np.logspace(-2, -0.8, 8)

        reflected_intensity = unumpy.uarray(y, dy)

        exp_y = unumpy.uarray(y, dy) / 1e9

        reflected_intensity = stitching.normalise_ter(x, reflected_intensity)

        assert_almost_equal(
            unumpy.nominal_values(reflected_intensity), unumpy.nominal_values(exp_y)
        )
        assert_almost_equal(
            unumpy.std_devs(reflected_intensity),
            [
                0.008165,
                0.008165,
                0.008165,
                0.0036489,
                0.0014896,
                0.0007217,
                0.0003891,
                0.0002309,
            ],
        )

    def test_normalise_ter_a(self):
        """
        Test normalisation to 1 where only one point is before cut-off
        """
        y = [1e9, 1e9, 1e9, 3.16e8, 1.29e8, 6.25e7, 3.37e7, 2.00e7]
        dy = [1e7, 1e7, 1e7, 3.16e6, 1.29e6, 6.25e5, 3.37e5, 2.00e5]

        x = np.logspace(-1.5, 4, 8)

        reflected_intensity = unumpy.uarray(y, dy)

        exp_y = unumpy.uarray(y, dy) / 1e9

        reflected_intensity = stitching.normalise_ter(x, reflected_intensity)

        assert_almost_equal(
            unumpy.nominal_values(reflected_intensity), unumpy.nominal_values(exp_y)
        )
        assert_almost_equal(
            unumpy.std_devs(reflected_intensity),
            [
                0.0,
                0.0141421,
                0.0141421,
                0.0044689,
                0.0018243,
                0.0008839,
                0.0004766,
                0.0002828,
            ],
        )

    def test_normalise_ter_b(self):
        """
        Test normalisation to 1 where the gradient max is after the first point
        """
        y = [1e9, 1.29e8, 6.25e7, 3.37e7, 2.00e7]
        dy = [1e7, 1.29e6, 6.25e5, 3.37e5, 2.00e5]

        x = np.logspace(-2, -0.8, 5)

        reflected_intensity = unumpy.uarray(y, dy)

        exp_y = unumpy.uarray(y, dy) / 1e9

        reflected_intensity = stitching.normalise_ter(x, reflected_intensity)

        assert_almost_equal(
            unumpy.nominal_values(reflected_intensity), unumpy.nominal_values(exp_y)
        )
        assert_almost_equal(
            unumpy.std_devs(reflected_intensity),
            [0.0, 0.0018243, 0.0008839, 0.0004766, 0.0002828],
        )

    def test_rebin(self):
        """
        Test rebin with specific q
        """
        reflected_intensity = unumpy.uarray(
            np.arange(1, 9, 1), np.arange(1, 9, 1) * 0.1
        )
        q_vectors = unumpy.uarray([1, 2, 3, 4, 4, 5, 6, 7], np.zeros(8))

        exp_y = [1, 2, 3, 4.5, 6, 7]
        new_x = [1, 2, 3, 4, 5, 6, 7]
        exp_x = [1, 2, 3, 4, 5, 6]

        q, r = stitching.rebin(q_vectors, reflected_intensity, new_q=new_x)

        assert_almost_equal(unumpy.nominal_values(r), exp_y)
        assert_almost_equal(unumpy.nominal_values(q), exp_x)

    def test_rebin_interpolate(self):
        """
        Test rebin with specific q with interpolate
        """
        reflected_intensity = unumpy.uarray(
            np.linspace(1, 9, 5), np.linspace(1, 9, 5) * 0.01
        )
        q_vectors = unumpy.uarray([1, 3, 5, 7, 9], np.zeros(5))

        exp_y = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        new_x = [1, 2, 3, 4, 5, 6, 7, 8, 9]

        q, r = stitching.rebin(q_vectors, reflected_intensity, new_q=new_x,
                               interpolate=True)

        assert_almost_equal(unumpy.nominal_values(r), exp_y)

    def test_rebin_a(self):
        """
        Test rebin with default
        """
        reflected_intensity = unumpy.uarray(
            np.arange(1, 9, 1), np.arange(1, 9, 1) * 0.1
        )
        q_vectors = unumpy.uarray([1, 2, 3, 4, 4, 5, 6, 7], np.zeros(8))

        exp_y = [1, 2, 3, 4.5, 6, 7]

        q, r = stitching.rebin(q_vectors, reflected_intensity)

        assert_almost_equal(unumpy.nominal_values(r), exp_y)



from islatu import background
from islatu import corrections
from islatu import cropping
from islatu import image
from islatu import io
from islatu import refl_data
from islatu import stitching
from yaml import load
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
import datetime
from os import path
from ast import literal_eval as make_tuple
from uncertainties import ufloat
import numpy as np

function_map = {'gaussian_1d': background.fit_gaussian_1d,
                'guassian_2d': background.fit_gaussian_2d,
                'i07': io.i07_dat_parser,
                'crop': cropping.crop_2d,
                'crop_peak': cropping.crop_around_peak_2d,
                }

class Foreperson:
    def __init__(self, metadata):
        self.parser = None
        self.visit_id = None
        self.year = datetime.datetime.now().year
        self.directory_path = None
        self.crop_function = function_map['crop_peak']
        self.crop_kwargs = None
        self.bkg_function = function_map['gaussian_1d']
        self.bkg_kwargs = None
        self.dcd_normalisation = None
        self.sample_size = None
        self.beam_width = None
        self.n_qvectors = 50
        self.q_min = None
        self.q_max = None
        self.q_step = None
        self.q_shape = 'linear'
        self.n_columns = 3
        self.theta_axis_name = "dcdtheta"
        self.q_axis_name = "qdcd"
        self.transpose = False
        self.qz_dimension = 1
        self.qxy_dimension = 0
        self.pixel_max = 1e6
        self.hot_pixel_max = 1e5
        self.metadata = metadata
        self.creator = {'name': 'Unknown',
                        'affiliation': 'Unknown',
                        'time': datetime.datetime.now().isoformat()}
        self.data_source = {'experiment': {'probe': 'X-rays',
                                           'instrument': 'i07',
                                           'sample': "Unknown"}}
        self.reduction = {'software': 'islatu', 'yaml': None}
        if self.n_columns == 3:
            self.data = {'column 1': 'Qz / Aa^-1', 'column 2': 'RQz', 'column 3': 'sigma RQz , standard deviation'}
        elif self.n_columns == 4:
            self.data = {'column 1': 'Qz / Aa^-1', 'column 2': 'RQz', 'column 3': 'sigma RQz , standard deviation', 'column 4': 'dQz-by-Qz , resolution'}
        if metadata is not None:
            if 'creator' in metadata.keys():
                self.creator = metadata['creator']
            if 'data source' in metadata.keys():
                self.data_source = metadata['data source']
        self.header = None


    def ready(self, directory):
        directory_path = directory.format(self.data_source['experiment']['instrument'], self.year, self.visit_id)
        if path.isdir(directory_path):
            self.directory_path = directory_path
        else:
            raise FileNotFoundError("The experiment directory cannot be found.")
        self.header = ' creator:\n   name:'+self.creator['name']+'\n'\
                      '   affiliation:'+self.creator['affiliation']+'\n'\
                      '   time:'+self.creator['time']+'\n data source:\n'\
                      '   experiment:\n     '\
                      'probe:'+self.data_source['experiment']['probe']+'\n     '\
                      'instrument:'+self.data_source['experiment']['instrument']+'\n     '\
                      'sample:'+self.data_source['experiment']['sample']+'\n '\
                      'reduction:\n   software:'+self.reduction['software']+'\n   '\
                      'yaml:'+self.reduction['yaml']+'\n data:\n   '\
                      'column 1:'+self.data['column 1']+'\n   '\
                      'column 2:'+self.data['column 2']+'\n   '\
                      'column 3:'+self.data['column 3']
        if self.n_columns == 4:
            self.header += '\n   column 4:'+self.data['column 4']

def reduce(run_numbers, yaml_file, directory='/dls/{}/data/{}/{}/', metadata=None):
    """
    The runner that parses the yaml file and performs the data reduction.

    run_numbers (:py:attr:`list` of :py:attr:`int`): Reflectometry scans that
        make up the profile.
    yaml_file (:py:attr:`str`): File path to instruction set.
    metadata (:py:attr:`dict`): If metadata should be included in the outpuit
        file, pass this as a :py:attr:`dict`.
    """
    the_boss = Foreperson(metadata)
    y_file = open(yaml_file, 'r')
    recipe = load(y_file, Loader=Loader)
    y_file.close()
    the_boss.reduction['yaml'] = yaml_file
    keys = recipe.keys()
    if 'instrument' in keys:
        the_boss.data_source['experiment']['instrument'] = recipe['instrument']
        the_boss.parser = function_map[recipe['instrument']]
    else:
        raise ValueError("No instrument given in {}.".format(yaml_file))
    if 'visit' in keys:
        the_boss.visit_id = recipe['visit']['visit_id']
        if 'year' in recipe['visit'].keys():
            the_boss.year = recipe['visit']['year']
    else:
        raise ValueError("No visit given in {}.".format(yaml_file))
    if 'crop' in keys:
        the_boss.crop_function = function_map[recipe['crop']['method']]
        if 'kwargs' in recipe['crop']:
            the_boss.crop_kwargs = recipe['crop']['kwargs']
    if 'background' in keys:
        the_boss.bkg_function = function_map[recipe['background']['method']]
        if 'kwargs' in recipe['background']:
            the_boss.bkg_kwargs = recipe['background']['kwargs']
    if 'setup' in keys:
        if 'dcd_normalisation' in recipe['setup'].keys():
            the_boss.dcd_normalisation = recipe['setup']['dcd_normalisation']
        if 'sample_size' in recipe['setup'].keys():
            the_boss.sample_size = make_tuple(recipe['setup']['sample_size'])
            try:
                _ = len(the_boss.sample_size)
                the_boss.sample_size = ufloat(the_boss.sample_size[0], the_boss.sample_size[1])
            except TypeError:
                pass
        else:
            raise ValueError("No sample_size given in setup of {}.".format(yaml_file))
        if 'beam_width' in recipe['setup'].keys():
            the_boss.beam_width = make_tuple(recipe['setup']['beam_width'])
            try:
                _ = len(the_boss.beam_width)
                the_boss.beam_width = ufloat(the_boss.sample_size[0], the_boss.sample_size[1])
            except TypeError:
                pass
        else:
            raise ValueError("No beam_width given in setup of {}.".format(yaml_file))
        if 'theta_axis' in recipe['setup'].keys():
            the_boss.theta_axis_name = recipe['setup']['theta_axis']
        if 'q_axis' in recipe['setup'].keys():
            the_boss.q_axis_name = recipe['setup']['q_axis']
        if 'transpose' in recipe['setup'].keys():
            the_boss.transpose = recipe['setup']['transpose']
            if the_boss.transpose:
                the_boss.qz_dimension = 0
                the_boss.qxy_dimension = 1
        if 'pixel_max' in recipe['setup'].keys():
            the_boss.pixel_max = recipe['setup']['pixel_max']
        if 'hot_pixel_max' in recipe['setup'].keys():
            the_boss.hot_pixel_max = recipe['setup']['hot_pixel_max']
    else:
        raise ValueError("No setup given in {}.".format(yaml_file))
    if 'rebin' in keys:
        if 'n_qvectors' in recipe['rebin'].keys():
            the_boss.n_qvectors = recipe['rebin']['n_qvectors']
        elif 'min' in recipe['rebin'].keys() and 'max' in recipe['rebin'].keys() and 'step' in recipe['rebin'].keys():
            the_boss.q_min = recipe['rebin']['min']
            the_boss.q_max = recipe['rebin']['max']
            the_boss.q_step = recipe['rebin']['step']
            if 'shape' in recipe['rebin'].keys():
                the_boss.q_shape = recipe['rebin']['shape']
        else:
            raise ValueError("Please define parameters of rebin in {}.".format(yaml_file))
    if 'output_columns' in keys:
        the_boss.output_columns = recipe['output_columns']

    the_boss.ready(directory)

    files_to_reduce = []
    for f in run_numbers:
        files_to_reduce.append(the_boss.directory_path + f + '.dat')


    print("-" * 10)
    print('File Parsing')
    print("-" * 10)
    refl = refl_data.Profile(files_to_reduce, the_boss.parser, the_boss.q_axis_name,
                             the_boss.theta_axis_name, None, 0, the_boss.pixel_max,
                             the_boss.hot_pixel_max, the_boss.transpose)
    print("-" * 10)
    print('Cropping')
    print("-" * 10)
    refl.crop(the_boss.crop_function, the_boss.crop_kwargs)

    print("-" * 10)
    print('Background Subtraction')
    print("-" * 10)
    refl.bkg_sub(the_boss.bkg_function, the_boss.bkg_kwargs)

    print("-" * 10)
    print('Estimating Resolution Function')
    print("-" * 10)
    refl.resolution_function(the_boss.qz_dimension, progress=True)

    print("-" * 10)
    print('Performing Data Corrections')
    print("-" * 10)
    if the_boss.dcd_normalisation is not None:
        itp = corrections.get_interpolator(the_boss.dcd_normalisation, the_boss.parser)
        refl.qdcd_normalisation(itp)
    refl.footprint_correction(the_boss.beam_width, the_boss.sample_size)
    refl.transmission_normalisation()
    refl.concatenate()
    refl.normalise_ter()

    print("-" * 10)
    print('Rebinning')
    print("-" * 10)
    if the_boss.q_min is None:
        refl.rebin(number_of_q_vectors=the_boss.n_qvectors)
    else:
        if the_boss.q_space == 'linear':
            spacing = np.linspace
        elif the_boss.q_space == 'log':
            spacing = np.logspace
        refl.rebin(new_q=spacing(the_boss.q_min, the_boss.q_max, the_boss.q_step))

    if the_boss.output_columns == 3:
        data = np.array([refl.q, refl.R, refl.dR]).T
        np.savetxt(the_boss.directory_path + '/processing/XRR_' + run_numbers[0] + '.dat', data, header='{}'.format(the_boss.header))
    elif the_boss.output_columns == 4:
        data = np.array([refl.q, refl.R, refl.dR, refl.dq]).T
        np.savetxt(the_boss.directory_path + '/processing/XRR_' + run_numbers[0] + '.dat', data, header='{}'.format(the_boss.header))
    print("-" * 10)
    print('Reduced Data Stored in Processing Directory')
    print("-" * 10)



"""
Tests for corrections module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)
# pylint: disable=R0201

import os
import unittest
import numpy as np
import io
import islatu
from PIL import Image as PILIm
from numpy.testing import assert_almost_equal, assert_equal
from uncertainties import unumpy as unp
from islatu.image import Image
from islatu import cropping, background, image

EXAMPLE_FILE = (
    "0.   0.   1.   1.   4. 113. 117.   7.   1.   0. \n"
    "0.   0.   0.   3.   4. 127. 144.   9.   2.   0. \n"
    "2.   0.   0.   7.   7. 232. 271.  13.   5.   2. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "2.0000e+00 2.0000e+00 5.0000e+00 1.3000e+01 "
    "3.5800e+02 3.4490e+04 3.1763e+04 9.1100e+02 "
    "5.5000e+01 7.0000e+00 \n"
    "2.0000e+00 2.0000e+00 9.0000e+00 2.0000e+01 "
    "9.6300e+02 6.5535e+04 5.5515e+04 1.4450e+03 "
    "8.2000e+01 1.0000e+01 \n"
    "2.000e+00 2.000e+00 3.000e+00 2.100e+01 "
    "1.080e+02 5.337e+03 3.077e+03 1.900e+02 "
    "2.500e+01 8.000e+00 \n"
    "0.   2.   1.   2.  27. 697. 324.  25.   6.   0. \n"
    "0.   2.   1.   3.  16. 525. 245.  15.   4.   3. \n"
    "0.   0.   0.   1.   4. 355. 167.   4.   1.   0."
)

EXAMPLE_HOT_PIXEL = (
    "5.   0.   1.   1.   4. 113. 117.   7.   1.   0. \n"
    "0.   0.   0.   3.   4. 127. 144.   9.   2.   0. \n"
    "2.   0.   5.000e+04   7.   7. 232. 271.  13.   5.   2. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "0.   0.   0.   1.   4. 355. 167.   4.   1.   0."
)

EXAMPLES_HOT_PIXEL_CORNERA = (
    "5.000e+04   2.   1.   1.   4. 113. 117.   7.   1.   0. \n"
    "5.   3.   0.   3.   4. 127. 144.   9.   2.   0. \n"
    "2.   0.   5.   7.   7. 232. 271.  13.   5.   2. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "0.   0.   0.   1.   4. 355. 167.   4.   1.   0."
)

EXAMPLES_HOT_PIXEL_CORNERB = (
    "5.   0.   1.   1.   4. 113. 117.   7.   1.   0. \n"
    "0.   0.   0.   3.   4. 127. 144.   9.   2.   0. \n"
    "2.   0.   5.   7.   7. 232. 271.  13.   5.   2. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  10.   3. \n"
    "1.   0.   5.   6.  31. 672. 703.  55.  9.   3. \n"
    "0.   0.   0.   1.   4. 355. 167.   4.   0.   5.0000e+04"
)


class TestImage(unittest.TestCase):
    """
    Unit tests for Image class
    """

    def test_init(self):
        """
        Test file reading
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        data = io.StringIO(EXAMPLE_FILE)
        expected_image = np.loadtxt(data)
        assert_equal((10, 10), test_image.shape)
        assert_almost_equal(expected_image, test_image.n)

    def test_hot_pixel(self):
        """
        Test file reading
        """
        b = io.StringIO(EXAMPLE_HOT_PIXEL)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf, hot_pixel_max=4e4)
        assert_equal((10, 10), test_image.shape)
        assert_almost_equal(2, test_image.n[2, 2])

    def test_hot_pixel_cornera(self):
        """
        Test file reading
        """
        b = io.StringIO(EXAMPLES_HOT_PIXEL_CORNERA)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf, hot_pixel_max=4e4)
        assert_equal((10, 10), test_image.shape)
        assert_almost_equal(2, test_image.n[0, 0])

    def test_hot_pixel_cornerb(self):
        """
        Test file reading
        """
        b = io.StringIO(EXAMPLES_HOT_PIXEL_CORNERB)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf, hot_pixel_max=4e4)
        assert_equal((10, 10), test_image.shape)
        assert_almost_equal(4, test_image.n[-1, -1])

    def test_init_with_transpose(self):
        """
        Test for transposing with reading
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf, transpose=True)
        data = io.StringIO(EXAMPLE_FILE)
        expected_image = np.loadtxt(data, unpack=True)
        assert_equal((10, 10), test_image.shape)
        assert_almost_equal(expected_image, test_image.n)

    def test_nominal_values(self):
        """
        Test nominal values
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        data = io.StringIO(EXAMPLE_FILE)
        expected_image = np.loadtxt(data)
        assert_equal((10, 10), test_image.shape)
        assert_almost_equal(expected_image, test_image.n)
        assert_almost_equal(expected_image, test_image.nominal_values)

    def test_std_devs(self):
        """
        Test standard devs
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        data = io.StringIO(EXAMPLE_FILE)
        load = np.loadtxt(data)
        expected_image = np.sqrt(load)
        expected_image[np.where(load == 0)] = 1
        assert_equal((10, 10), test_image.shape)
        assert_almost_equal(expected_image, test_image.std_devs)
        assert_almost_equal(expected_image, test_image.s)

    def test_repr(self):
        """
        Test repr
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        data = io.StringIO(EXAMPLE_FILE)
        load = np.loadtxt(data)
        expected_image_e = np.sqrt(load)
        expected_image_e[np.where(load == 0)] = 1
        assert_almost_equal(load, unp.nominal_values(test_image.__repr__()))
        assert_almost_equal(expected_image_e, unp.std_devs(test_image.__repr__()))

    def test_str(self):
        """
        Test str
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        data = io.StringIO(EXAMPLE_FILE)
        load = np.loadtxt(data)
        expected_image_e = np.sqrt(load)
        expected_image_e[np.where(load == 0)] = 1
        assert_almost_equal(load, unp.nominal_values(test_image.__str__()))
        assert_almost_equal(expected_image_e, unp.std_devs(test_image.__str__()))

    def test_crop(self):
        """
        Test crop
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        test_image.crop(cropping.crop_around_peak_2d, x_size=2, y_size=2)
        assert_equal((2, 2), test_image.shape)

    def test_background_subtraction(self):
        """
        Test background_subtraction
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        test_image.background_subtraction(background.fit_gaussian_2d)
        assert_equal((10, 10), test_image.shape)

    def test_sum(self):
        """
        Test sum
        """
        b = io.StringIO(EXAMPLE_FILE)
        buf = io.BytesIO()
        im = PILIm.fromarray(np.loadtxt(b).astype(np.uint32))
        im.save(buf, format="png")
        buf.seek(0)
        test_image = Image(buf)
        assert_equal(isinstance(test_image.sum().n, float), True)
        assert_equal(isinstance(test_image.sum().s, float), True)
